.loadPackage <- function(pkg){
  
  if(length(grep(pkg, search())) >0 ){
    return(invisible(TRUE))
  }
  
  result <- tryCatch(suppressPackageStartupMessages(
    require(pkg, quietly=TRUE, character.only=TRUE)
  ))
  if (!result){
    stop(paste("Please install package",pkg))
  }
  return(invisible(result))
}


.checkInputs <-
  function(location,
           groups) {
    ## check that it is not null
    if (is.null(location)) {
      stop("Please indicate the location to the interaction files")
    }
    
    ## check that it is a list
    if (class(location) != "list") {
      stop("the location paramter should be a list")
    }
    
    ## check that the groups are provided
    if (is.null(groups)) {
      stop("No groups are indicated")
    }
    
    ## check that the lengths of the groups equals the number of samples
    if (length(location) != length(groups)) {
      stop("the length of provided groups does not match the number of provided samples")
    }
    
    
    sampPerGroup <- table(groups)
    
    ## check that we have two groups
    
    if (length(sampPerGroup) > 2) {
      stop("This version of FIND does not support more than two condition comparison\n")
    }
    
    ## check that we have at least two replicate in each group
    
    if (min(sampPerGroup) < 2) {
      stop("Each condition should at least have two replicates")
    }
    
  }

.checkHicExperiment <- function(hicExpr){
  

  
  
  groups = hicExpr@groups
  ## check that the groups are provided
  if (is.null(groups)) {
    stop("No groups are indicated")
  }
  
  ## check that the lengths of the groups equals the number of samples
  if (length(hicExpr@hic_matrices) != length(groups)) {
    stop("the length of provided groups does not match the number of provided samples")
  }
  
  
  sampPerGroup <- table(groups)
  
  ## check that we have two groups
  
  if (length(sampPerGroup) > 2) {
    stop("This version of FIND does not support more than two condition comparison\n")
  }
  
  ## check that we have at least two replicate in each group
  
  if (min(sampPerGroup) < 2) {
    stop("Each condition should at least have two replicates")
  }
  
}

.checkInputs2 <-
  function(location,
           groups,
           type = c("juicebox", "tab")) {
    ## check that it is not null
    if (is.null(location)) {
      stop("Please indicate the location of the hic files")
    }
    
    ## check that it is a list
    if (class(location) != "list") {
      stop("the location paramter should be a list")
    }
    
    ## check that the groups are provided
    if (is.null(groups)) {
      stop("No groups are indicated")
    }
    
    ## check that the lengths of the groups equals the number of samples
    if (length(location) != length(groups)) {
      stop("the length of provided groups does not match the number of provided samples")
    }
    
    
    sampPerGroup <- table(groups)
    
    ## check that we have two groups
    
    if (length(sampPerGroup) > 2) {
      stop("This version of FIND does not support more than two condition comparison\n")
    }
    
    ## check that we have at least two replicate in each group
    
    if (min(sampPerGroup) < 2) {
      stop("Each condition should at least have two replicates")
    }
    
    
    type <- match.arg(type)
    
    
    ## check that the path to the provided files is correct.
    
    if (type == "juicebox") {
      for (f in location) {
        if (!file.exists(f)) {
          stop(paste("Could not find file", f))
        }
        ## check that it is a valid .hic file
        checkFormat(f)
      }
    }
    else{
      for (f in location) {
        # check that the directory exists
        if (!dir.exists(f)) {
          stop("Could not find the directory", f)
        }
        
        ## check if is not empty
        fls <- list.files(f)
        if (length(fls) == 0) {
          stop(paste("It seems that the directory", f, "is empty"))
        }
        ## we don't check the format of the files for the momenent only when needed
      }
    }
  }


.checkMatrices <- function(Hic1, Hic2) {
  ## check that the user provides a list (redudunant, to avoid problems)
  if (class(Hic1) != "list") {
    stop("Hic1 sould be a list")
  }
  
  if (class(Hic2) != "list") {
    stop("Hic2 sould be a list")
  }
  
  if (length(Hic1) == 0) {
    stop("Hic1 is empty")
  }
  
  if (length(Hic2) == 0) {
    stop("Hic2 is empty")
  }
  
  ## check that that these are list of matrices
  for (i in 1:length(Hic1)) {
    if (class(Hic1[[i]]) != "dgCMatrix") {
      stop(message("The", i, "-th matrix in Hic1 is not dgCMatrix"))
    }
  }
  
  for (i in 1:length(Hic2)) {
    if (class(Hic2[[i]]) != "dgCMatrix") {
      stop(message("The", i, "-th matrix in Hic2 is not dgCMatrix"))
    }
  }
  
  
  ## check if the provided matrices have the same dimensions
  
  matDims <- dim(Hic1[[1]])
  
  for (i in 2:length(Hic1)) {
    if (!all(dim(Hic1[[i]]) == matDims)) {
      stop(message("The", i, "-th in Hic1 does't have the same dimension"))
    }
  }
  
  
  for (i in 1:length(Hic2)) {
    if (!all(dim(Hic2[[i]]) == matDims)) {
      stop(message("The", i, "-th in Hic2 does't have the same dimension"))
    }
  }
}

# setMethod("show", "HicDCI",
#           function(object) {
#             grps <- factor(object@groups)
#             
#             grps_list <- list()
#             
#             for (i in 1:length(levels(grps))) {
#               grps_list[[i]] <- which(object@groups == levels(grps)[i])
#             }
#             
#             cat(paste("We have", length(levels(grps)), "conditions\n"))
#             cat(paste("Format:", object@type, "\n"))
#             for (i in 1:length(levels(grps))) {
#               cat("Condition", i, "has", length(grps_list[[i]]), "replicates\n")
#             }
#           })



#' Loading the Hi-C replicates from the two conditions 
#'
#' @param locations a list that contains the path to the Hi-C matrices of the two conditions.
#' Each matrix should be represented in a 3-columned format: bin1ID bin2ID  Freq.
#' @param binIndices path to the file containing the bins genomic location. 
#' If you use HiC-Pro, you can the "abs.bed" file. Otherwise, you can provide a 
#' 4-columned file that contains the chromosome, start, end and binID.
#' @param groups numeric. A numeric vector that indicates the group of each sample (For the moment only two groups are allowed)
#' @param type character. The format in which the Hi-C data is encoded. 
#' For the moment we support the HiC-Pro format, probably we will addd new formats 
#' such as .hic format.
#'
#' @return a HicExperiment object
#' @export
#'
#' @examples
#' \dontrun{
#'   hic1 <-
#' }
setMethod("loadHicExperiment", signature = c(
  locations = "list",
  binIndices="character",
  type="character",
  groups = "numeric",
  genome = "character"
),
function(locations = NULL,
         binIndices = NULL,
         type = c("HicPro"),
         groups=NULL,
         genome=NULL) {
  ## check if the inputs are corretly provided
  .checkInputs(locations, groups)
  
  type <- match.arg(type)
  
  
  hic_matrices <- list()
  
  for(i in 1:length(locations)){
    message(paste("Loading:",locations[[i]]))
    
    hic_matrices[[i]] <- importC(locations[[i]],
                                 xgi.bed = binIndices,
                                 allPairwise = F,
                                 rm.trans = T)  
  }
  
  names(hic_matrices) <- names(locations)
  ## Get the list of common Hi-C contract maps between replicates
  ## N.B: Just in case someone gives different set of chromomosomes in
  ## different conditions.
  
  
  common_interactions <- names(hic_matrices[[1]])
  
  for(i in 2:length(hic_matrices)){
    common_interactions <- intersect(common_interactions,  names(hic_matrices[[i]]))
  }
  
  # remove the un-common ones
  
  for(i in 1:length(hic_matrices)){
    hic_matrices[[i]] <- hic_matrices[[i]][common_interactions]
  }
  
  hicCol <- new("HicExperiment",
                hic_matrices= hic_matrices, 
                groups=groups, 
                DCIs=list(), 
                genome=genome)
  return(hicCol)
})









#' Read a juicebox formatted Hi-C matrix
#'
#' @param filePath character. Path to the .hic file.
#' @param chromosome character. The chromosome we are intersted to read.
#' @param normMethod character. The normalization method to use.
#' Only the ones accepted by juicebox can be used. The accepted values are : "NONE","VC","VC_SQRT" or "KR"
#' @param resolution numeric. The resolution to use. Only the resolutions supported by juicebox are used.
#'
#' @return A sparse matrix of class "dgCMatrix" with only the upper triangle filled.
#' For the moment this function works only on linux machines.
#' @export
#'
#' @examples
setMethod("readJuiceBoxFormat", signature = c(
  filePath = "character",
  chromosome = "character",
  normMethod = "character",
  resolution = "numeric"
),
function(filePath,
         chromosome,
         normMethod = c("NONE", "VC", "VC_SQRT", "KR"),
         resolution = 5000) {
  
  filePath <- normalizePath(filePath)
  chromosome <- gsub("chr", "", chromosome, ignore.case = T)
  
  normMethod <- match.arg(normMethod)
  
  cmd <-
    paste(normMethod,
          filePath,
          chromosome,
          chromosome,
          "BP",
          resolution)
  print(cmd)
  hicMap <- straw_R(cmd)
  hicMap$x <-
    as.numeric(as.character(hicMap$x)) / resolution + 1
  hicMap$y <-
    as.numeric(as.character(hicMap$y)) / resolution + 1
  
  m <- max(hicMap$y, hicMap$x)
  
  ## Remove NaN and Inf regions
  hicMap$counts <- as.numeric(as.character(hicMap$counts))
  pos <- which(!is.nan(hicMap$counts))
  hicMap <- hicMap[pos, ]
  pos <- which(!is.infinite(hicMap$counts))
  hicMap <- hicMap[pos, ]
  
  ## Convert to a SparceMatrix
  hicMap <-
    sparseMatrix(
      i = hicMap$x,
      j = hicMap$y,
      x = hicMap$counts,
      dims = c(m, m)
    )
  
  ## Take only the upper triangle
  hicMap <- as(triu(hicMap), "dgCMatrix")
  
  hicMap
})




setMethod("readTabformat", signature = c(
  filePath = "character",
  chromosome = "character",
  resolution = "numeric"
),
function(filePath, chromosome, resolution) {
  hicMap <- fread(filePath)
  
  if (nrow(hicMap) != 7) {
    stop(
      "Please verify that the interactions are in the format\nchr1\tstart1\tend1\tchr2\tstart2\tend2\tfreq"
    )
  }
  
  
  setnames(
    hicMap,
    colnames(hicMap),
    c("chr1", "start1", "end1", "chr2", "start2", "end2", "freq")
  )
  
  ## Consider only the intra-chromosomal interactions of the chromosome of interest
  hicMap <-
    subset(hicMap, chr1 == chromosome & chr2 == chromosome)
})




# setMethod("show",'HicCollection',
#           function(object){
#
#             nbFiles <- length(object@location)
#             nbReps <- table(object@groups)
#
#             cat("HicCollection object with :\n")
#             cat(paste("-)",nbFiles,"Hi-C matrices of type:",object@type,"\n"))
#             cat(paste("-)",nbReps[1]," Hi-C replicates in group 1 and",nbReps[2],"Hi-C replicates in group 2\n"))
#           })




## This routine will check if the given chromosome exists
## if the Hi-C matrix is in the juicebox format the check is straight forward.
## if the Hi-C matrix is in the tab format it should be located at
## object@location/chromosome.bed
## it should also be in the 3 columns format : bin1 bin2  frequency

.checkChromsome <- function(object, chromosome) {
  if (object@type == "juicebox") {
    
  } else{
    fname <- paste0(chromosome, ".bed")
    
    fname <- file.path(object@location, fname)
    
    if (!file.exists(fname)) {
      stop(paste0("Could not find file:\n", fname))
    }
    
    ## read the first 100 lines to check if the format is correct
  }
}


.getDCis <- function(res, method){
  
}




#' Detection of DCIs given a list of matrices
#'
#' @param Hic1 list of \code{dgCMatrix} matrices. Contains the Hi-C replicates for the 1st condition.
#' @param Hic2 list of \code{dgCMatrix} matrices. Contains the Hi-C replicates for the 2nd condition.
#' @param windowSize numeric. The sliding window width.
#' @param method character. Type of method to use to select the q-value cutoff.
#' possible values are : \code{hardCutof} in which the cutoff is applied dirrectly,
#' and \code{quantreg} in which a quantile regression method is used.
#' @param alpha numeric. The percentage of the k-nearest neighbors required
#' to be significantly different in each sliding window.
#' @param qvalue numeric. cutoff q-value. Taken into considiration i
#' @param isrOP_qval logical.
#'
#' @return
#' @export
#'
#' @examples
setMethod("getDCIs_fromMat", signature = signature(
  Hic1 = "list",
  Hic2 = "list",
  windowSize = "numeric",
  method = "character",
  alpha = "numeric"
),
function(Hic1 = NULL,
         Hic2 = NULL,
         windowSize = 3,
         method = c("hardCutoff", "quantreg"),
         alpha = 0.6,
         qvalue = 1e-3,
         isrOP_qval = TRUE,
         quantile = 0.99,
         resolution = 5000) {
  ## check if the inputs are correct

  .checkMatrices(Hic1, Hic2)
  method = match.arg(method)
  message("detecting differential regions")
  
  res <-
    getDiff(
      Hic1 = Hic1,
      Hic2 = Hic2,
      window = windowSize,
      percent = alpha
    )
  
  message("Adjusting qvalues")
  ncol_max <- max(ncol(Hic1[[1]]), ncol(Hic1[[2]]),
                  ncol(Hic2[[1]]), ncol(Hic2[[2]]))
  
  
  nrow_max <- max(nrow(Hic1[[1]]), nrow(Hic1[[2]]),
                  nrow(Hic2[[1]]), nrow(Hic2[[2]]))
  
  
  mrg_qval <- .adjustPvalueByDiagonal(res)
  
  if (method == "hardCutof") {
    
    mrg_sig <- .hardCutoff(mrg_qval,qvalue=qvalue);
  
  } else{
    ## Use quantile regression
    warning("Quantile regression can be very slow using hard clusting instead")
    #mrg_sig <- .quantRegressionthereshold(mrg_qval, windowSize, quantile)
    mrg_sig <- .hardCutoff(mrg_qval,qvalue=qvalue);
  }
  
  ## use the corrected qvalues instead
  
  qvals_mat <- sparseMatrix(
                      i = mrg_sig$i,
                      j = mrg_sig$j,
                      x = mrg_sig$padj,
                      dims = c(nrow_max, ncol_max)
                    )
  res2 = .getclusteredRegions(qvals_mat, windowSize, alpha)
  res2[res2!=0] = 1
  ## Return results
  # res2 <-
  #   sparseMatrix(
  #     i = mrg_sig$i,
  #     j = mrg_sig$j,
  #     x = 1,
  #     dims = c(nrow_max, ncol_max)
  #   )
  
  

  params <- list(windowSize = windowSize, 
                 method = method,
                 alpha= alpha, 
                 qvalue = qvalue, 
                 isrOP_qval = isrOP_qval, 
                 quantile = quantile)
  
  output <- new("DCIs",
                qvals = qvals_mat,
                DCIPos=res2,
                params = params
                )
  
  return(output)
})




setMethod("getDCIs_fromSim", signature = signature(
  x= "SimulatedHiC",
  windowSize = "numeric",
  method = "character",
  alpha = "numeric"
),
function(x,
         windowSize = 3,
         method = c("hardCutoff", "quantreg"),
         alpha = 0.6,
         qvalue = 1e-3,
         isrOP_qval = TRUE,
         quantile = 0.99,
         resolution = 5000){
  
  getDCIs_fromMat(x@Hic1,x@Hic2,
                  windowSize,method,
                  alpha, qvalue, 
                  isrOP_qval,quantile,
                  resolution)
}
)


setMethod("getDCIs_parallel", signature = signature(
  Hic1 = "list",
  Hic2 = "list",
  windowSize = "numeric",
  method = "character",
  alpha = "numeric",
  quantile = "numeric",
  qvalue="numeric",
  isrOP_qval='logical',
  chunk_size = "numeric",
  nbProcessors= "numeric",
  verbose="logical"
  
), function(Hic1,Hic2, windowSize,
            method=c("hardCutoff", "quantreg"), 
            alpha=0.6,
            quantile=0.99,
            qvalue=1e-6,
            isrOP_qval=TRUE,
            chunk_size=500, 
            nbProcessors=2,
            verbose=TRUE){
  
  
  # check if the matrices have the same size 
  .checkMatrices(Hic1,Hic2)
  
  if(length(chunk_size)>2){
    stop("chunk_size can a 1 or a 2 dimensional vector")
  }
  
  if(length(chunk_size) ==1){
    chunk_size <- rep(chunk_size,2)
  }
  # create bigmemmory refrences
  nrows = nrow(Hic1[[1]])
  ncols = ncol(Hic1[[1]])
  
  # create big memory objects
  
  bigMems_hic1 <- list()
  bigMems_hic1_desc  <- list()
  
  bigMems_hic2 <- list()
  bigMems_hic2_desc <- list()
  
  message("Creating big.memory description files for Hic1 matrices\n")
  warning("More memory will be used in this step")
  for(i in 1:length(Hic1)){
    bigMems_hic1[[i]] <- as.big.matrix(x = as.matrix(Hic1[[i]]))
    bigMems_hic1_desc[[i]] <- describe(bigMems_hic1[[i]])
  }
  
  message("Creating big.memory description files for Hic2 matrices\n")
  warning("More memory will be used in this ste\n")
  for(i in 1:length(Hic2)){
    bigMems_hic2[[i]] <- as.big.matrix(x = as.matrix(Hic2[[i]]))
    bigMems_hic2_desc[[i]] <- describe(bigMems_hic2[[i]])
  }
  
  
  
  # process_chunk will be called to process each shunk
  
  process_chunk <- function(j,i, chunksize = chunk_size,
                            w=windowSize, 
                            alpha=alpha,
                            hic1_desc = NULL,
                            hic2_desc = NULL,
                            res_matric_desc=NULL){
    
    # check not NULL
    if(is.null(hic1_desc) | is.null(hic2_desc)){
      stop("You need to provide the descriptor of each matrix")
    }
    
    # check not empty
    if(length(hic1_desc) <2 | length(hic2_desc) <2){
      stop("At least two matrices in each condition should be provided")
    }
    
    # check that it is a big.matrix.descritor
    
    if(! all(sapply(hic1_desc, class) == "big.matrix.descriptor") |
       ! all(sapply(hic2_desc, class) == "big.matrix.descriptor")){
      stop("big.matrix.descriptor objects should be provided")
    }
    
    # attach matrices
    
    HIC1 = list()
    for(n in 1:length(hic1_desc)){
      HIC1[[n]] <- attach.big.matrix(hic1_desc[[n]])  
    }
    
    HIC2 = list()
    for(n in 1:length(hic2_desc)){
      HIC2[[n]] <- attach.big.matrix(hic2_desc[[n]])  
    }
    
    
    
    res_matrix = attach.big.matrix(res_matric_desc)
    
    nRow <- nrow(HIC1[[1]])
    nCol <- ncol(HIC1[[1]])
    
    gap <-floor(w/2)
    
    startRow <- (chunksize[1]) * (i-1) -gap
    if(startRow <= 0) startRow <- 1
    endRow <- (chunksize[1]) * i
    endRow <- ifelse(endRow > nRow, nRow, endRow)
    
    startCol <- (chunksize[2]) * (j-1) -gap
    if(startCol <= 0) startCol <- 1
    endCol <- (chunksize[2]) * j
    endCol <- ifelse(endCol > nCol, nCol, endCol)
    print(paste0("Processing region [",startRow,":",endRow,",", startCol,":",endCol,"]"))
    
    hic1= list()
    for(n in 1:length(HIC1)){
      hic1[[n]] <- as(HIC1[[n]][startRow:endRow, startCol:endCol],"dgCMatrix")
    }
    
    hic2= list()
    for(n in 1:length(HIC2)){
      hic2[[n]] <- as(HIC2[[n]][startRow:endRow, startCol:endCol], "dgCMatrix")
    }
    
    sum1 <- 0
    for(n in 1:length(hic1)){
      sum1 <- sum1 +sum(hic1[[n]])
    }
    
    sum2 <- 0
    for(n in 1:length(hic2)){
      sum2 <- sum2 +sum(hic2[[n]])
    }
    
    if(sum1 ==0 & sum2==0){
      return(0L)
    }
    
    # Do differential analysis
    res <-
      getDiff(
        Hic1 = hic1,
        Hic2 = hic2,
        window = w,
        percent = alpha
      )
    
    # free some memory
    rm(hic1,hic2)
    
    ## write results
    out_starti = ifelse(startRow == 1, startRow, startRow+gap)
    out_endi = ifelse(endRow == nRow, endRow, endRow-gap)
    
    out_startj = ifelse(startCol == 1, startCol, startCol+gap)
    out_endj = ifelse(endCol == nCol, endCol, endCol-gap)
    
    
    si = ifelse(startRow == 1, 1, 1+gap)
    ei = ifelse(endRow == nRow, nrow(res), nrow(res)-gap)
    
    sj = ifelse(startCol == 1, 1, gap+1)
    ej = ifelse(endCol == nCol, ncol(res), ncol(res)-gap)
    
    res_matrix[out_starti:out_endi,out_startj:out_endj] <- as.matrix(res[si:ei, sj:ej])
    return(0L)
  }
  
  # parallel processing
  nbParRow <- ceiling(nrow(Hic1[[1]])/chunk_size[1])
  nbParCol <- ceiling(ncol(Hic1[[1]])/chunk_size[2])
  
  if(verbose){
    message(paste0("Splitting Hi-C matrix into (", nbParRow," x ",nbParRow,
                   ") blocks of size",chunk_size))
  }
  
  diff_res <- big.matrix(nrow = nrow(Hic1[[1]]), ncol = ncol(Hic1[[1]]),
                         type="double")
  diff_res_desc = describe(diff_res)
  
  if(nbProcessors > 1){
    message(paste("Using",nbProcessors,"Processors"))
    cl <- makeCluster(nbProcessors)
    registerDoParallel(cl)  
  }else{
    message("Using only one processor")
    registerDoSEQ() 
  }
  
  for(i in 1:nbParRow){
    print(paste("****** i=",i,"******"))      
    foreach(j = 1:nbParCol,.packages = "bigmemory") %dopar% {
      # fname <- paste0("~/Documents/tmp/FIND/window_",i,"_",j,".txt")
      # sink(fname)
      # message("Test")
      # message(paste("chunk size:",chunk_size))
      # message(paste("windowSize:",windowSize))
      # message(paste("Alpha:", alpha))
      
      #tryCatch({
        process_chunk(j = j,i = i,
                      chunksize = chunk_size,
                      w = windowSize,
                      alpha = alpha,
                      hic1_desc = bigMems_hic1_desc,
                      hic2_desc = bigMems_hic2_desc,
                      res_matric_desc = diff_res_desc)
        
      # }, warning = function(w) {
      #   print(paste("Warning:",w))
      # }, error = function(e) {
      #   stop(paste("error:",e))
      # })
      
    }
    #sink()
  }
  
  if(nbProcessors > 1){
    stopCluster(cl)  
  }
  
  
  ## freeying some memory
  rm(bigMems_hic1,bigMems_hic2,bigMems_hic1_desc,bigMems_hic2_desc)
  gc()
  warning("This step many hit the memory ceiling")
  
  
  res <- Matrix(as.matrix(diff_res),sparse = TRUE)
  res <- as(res,"dgCMatrix")
  rm(diff_res,diff_res_desc)
  
  mrg_qval <- .adjustPvalueByDiagonal(res)
  rm(res)
  ## No that we the qvalues let's get the DCis
  if (method == "hardCutof") {
    message('Doing hard cutoff')
    mrg_sig <- .hardCutoff(mrg_qval, qvalue=qvalue);
    
  } else{
    ## Use quantile regression
    #mrg_sig <- .quantRegressionthereshold(mrg_qval, windowSize, quantile)
    warning("Quantile regression can be very slow using hard clusting instead")
    #mrg_sig <- .quantRegressionthereshold(mrg_qval, windowSize, quantile)
    mrg_sig <- .hardCutoff(mrg_qval,qvalue=qvalue);
  }
  

  qvals_mat <- sparseMatrix(
    i = mrg_sig$i,
    j = mrg_sig$j,
    x = mrg_sig$padj,
    dims = c(nrows, ncols)
  )
  
  ## Return the DCIs positions
  # res2 <-
  #   sparseMatrix(
  #     i = mrg_sig$i,
  #     j = mrg_sig$j,
  #     x = 1,
  #     dims = c(nrows, ncols)
  #   )
  message('Detecting clusted regions')
  res2 = .getclusteredRegions(qvals_mat, windowSize, alpha)
  res2[res2!=0] = 1
  
  params <- list(windowSize = windowSize, 
                 method = method,
                 alpha= alpha, 
                 qvalue = qvalue, 
                 isrOP_qval = isrOP_qval, 
                 quantile = quantile,
                 chunk_size=chunk_size, 
                 nbProcessors=nbProcessors)

  
  output <- new("DCIs",
                qvals = qvals_mat,
                DCIPos=res2,
                params = params)
  
  return(output)
}
)



setMethod("getDCIs_HicExpr", signature = signature(
  HicExpr="HicExperiment",
  windowSize = "numeric",
  method = "character",
  alpha = "numeric",
  quantile = "numeric",
  qvalue="numeric",
  isrOP_qval='logical',
  chunk_size = "numeric",
  nbProcessors= "numeric",
  verbose="logical"
  
), function(HicExpr, windowSize,
            method=c("hardCutoff", "quantreg"), 
            alpha=0.6,
            quantile=0.99,
            qvalue=1e-6,
            isrOP_qval=TRUE,
            chunk_size=500, 
            nbProcessors=2,
            verbose=TRUE){
  
  .checkHicExperiment(HicExpr)
  
  
  interactions <- names(HicExpr@hic_matrices[[1]])
  
  grps <- unique(HicExpr@groups)
  
  pos1 <- which(HicExpr@groups == grps[1])
  pos2 <- which(HicExpr@groups == grps[2])
  ## Process interactions sequencly
  
  DCIs <- list()
  
  for(inter in interactions){
    
    ## get the Hi-C matrices
    Hic1 <- list()
    Hic2 <- list()
    
    for(i in 1:length(pos1)){
      Hic1[[i]] <- intdata(HicExpr@hic_matrices[[pos1[i]]][[inter]])
    }
    
    for(i in 1:length(pos2)){
      Hic2[[i]] <- intdata(HicExpr@hic_matrices[[pos2[i]]][[inter]])
    }
    
    message("*******")
    message(paste("*******Processing block",inter,"*******\n"))
    message("*******")
    
    ## Run parallel processing
    DCIs[[inter]] = getDCIs_parallel(Hic1 = Hic1,
                                Hic2 = Hic2,
                                windowSize = windowSize,
                                alpha = alpha, 
                                chunk_size=chunk_size,
                                method = method,
                                quantile = quantile,
                                qvalue = qvalue,
                                isrOP_qval = isrOP_qval,
                                nbProcessors = nbProcessors,
                                verbose =  TRUE)
  }
 
  HicExpr@DCIs <- DCIs
  
  return(HicExpr)
}
)





#' compareWithTrueDCI: Compare predicted results with the ground truth 
#'
#' @param predictedDCI dgCMatrix. Binarry matrix containing the position of the predicted DCIs
#' @param trueDCI dgCMatrix. Binnary matrix containing the true position of the DCIs
#'
#' @return
#'  A plot in which the orange sequate indicate the false negative, 
#'  the blue sequares indicate the successfully predicted DCIs,
#'  the gree sequares indicate the false positives.
#' @export
#'
#' @examples
setMethod("compareWithTrueDCI", signature = c(predictedDCI = "dgCMatrix",
                                              trueDCI = "dgCMatrix"),
          function(predictedDCI, trueDCI) {
            
            if (!all(dim(predictedDCI) == dim(trueDCI))) {
              stop("The provided matrices have different sizes")
            }
            
            
            .loadPackage("rasterVis")
            .loadPackage("RColorBrewer")
            
            myTheme <- rasterTheme(region = c("#FFFFCC",
                                              brewer.pal("Spectral", n =
                                                           10)))
            
            levelplot(as.matrix(2 * predictedDCI + trueDCI), 
                      par.settings = myTheme)
          })
