###########################################################################################
## This simulation code is based on the code in :                                       ###
## http://130.60.190.4/robinson_lab/edgeR_robust/robust_simulation.R                    ###
## And the related paper : http://130.60.190.4/robinson_lab/edgeR_robust/pipeline.pdf   ###
## And:  https://doi.org/10.1093/nar/gku310                                             ###
##                                                                                      ###
###########################################################################################

.getDataset <-
  function(counts,
           drop.extreme.dispersion = 0.1,
           drop.low.lambda = TRUE) {
    ## this function generates NB parameters from real dataset ##
    ## it is low-level function of NBsim ##
    d <- DGEList(counts)
    d <- calcNormFactors(d)
    cp <- round(cpm(d, normalized.lib.sizes = TRUE), 1)
    if (drop.low.lambda)
      d <- d[rowSums(cp > 1) >= 2,]
    d$AveLogCPM <- log2(rowMeans(cpm(d, prior.count = 1e-5)))
    d <- estimateGLMCommonDisp(d)
    d <- estimateGLMTrendedDisp(d)
    d <- estimateGLMTagwiseDisp(d)
    dispersion <- d$tagwise.dispersion
    AveLogCPM <- d$AveLogCPM
    if (is.numeric(drop.extreme.dispersion))
    {
      bad <-
        quantile(dispersion, 1 - drop.extreme.dispersion, names = FALSE)
      ids <- dispersion <= bad
      AveLogCPM <- AveLogCPM[ids]
      dispersion <- dispersion[ids]
    }
    dataset.AveLogCPM <- AveLogCPM
    dataset.dispersion <- dispersion
    dataset.lib.size <- d$samples$lib.size
    dataset.nTags <- nrow(d)
    list(
      dataset.AveLogCPM = dataset.AveLogCPM,
      dataset.dispersion = dataset.dispersion,
      dataset.lib.size = dataset.lib.size,
      dataset.nTags = dataset.nTags
    )
  }



.NBSim.neig <-
  function(dataset,
           group,
           nTags = 10000,
           nlibs = length(group),
           fix.dispersion = NA,
           lib.size = NULL,
           drop.low.lambda = TRUE,
           drop.extreme.dispersion = 0.1,
           add.outlier = FALSE,
           outlierMech = c("S", "R", "M"),
           pOutlier = 0.1,
           min.factor = 1.5,
           max.factor = 10,
           pDiff = .1,
           pUp = .5,
           foldDiff = 3,
           name = NULL,
           save.file = FALSE,
           file = NULL,
           only.add.outlier = FALSE,
           verbose = TRUE)

  {
    ## NBsim generate simulated count from the real dataset followed by the NB model ##
    require(edgeR)
    group = as.factor(group)
    
    sample.fun <- function(object)
    {
      ## it is low-level function of NBsim ##
      ## it samples from the real dataset ##
      nlibs <- object$nlibs
      nTags <- object$nTags
      AveLogCPM <- object$dataset$dataset.AveLogCPM
      dispersion <- object$dataset$dataset.dispersion
      
      id_r <- sample(length(AveLogCPM), nTags, replace = TRUE)
      Lambda <- 2 ^ (AveLogCPM[id_r])
      Lambda <- Lambda / sum(Lambda)
      Dispersion <- dispersion[id_r]
      id_0 <- Lambda == 0
      Lambda <- Lambda[!id_0]
      Dispersion <- Dispersion[!id_0]
      Lambda <- expandAsMatrix(Lambda, dim = c(nTags, nlibs))
      object$Lambda <- Lambda
      if (!is.na(fix.dispersion))
        Dispersion <-
        expandAsMatrix(fix.dispersion, dim = c(nTags, nlibs))
      else
        Dispersion <- expandAsMatrix(Dispersion, dim = c(nTags, nlibs))
      object$Dispersion <- Dispersion
      object
      
    }
    diff.fun <- function(object)
    {
      #browser()
      ## it is low-level function of NBsim ##
      ## it creates diff genes according to foldDiff ##
      group <- object$group
      pDiff <- object$pDiff
      pUp <-  object$pUp
      foldDiff <- object$foldDiff
      Lambda <- object$Lambda
      nTags <- object$nTags
      g <- group == levels(group)[1]
      ind <- sample(nTags, floor(pDiff * nTags))
      if (length(ind) > 0 & !foldDiff == 1) {
        fcDir <-
          sample(c(-1, 1),
                 length(ind),
                 prob = c(1 - pUp, pUp),
                 replace = TRUE)
        
        old_ind <- ind
        bin1 <- as.numeric(gsub("_\\d+", "", object$gene.names))
        bin2 <- as.numeric(gsub("\\d+_", "", object$gene.names))
        size <- max(c(bin1, bin2))
        
        Lambda[ind, g] <- Lambda[ind, g] * exp(log(foldDiff) / 2 * fcDir)
        Lambda[ind, !g] <-
          Lambda[ind, !g] * exp(log(foldDiff) / 2 * (-fcDir))
        dbg <- TRUE
        ## modify the neighors
        for (n in old_ind) {
          i <- bin1[n]
          j <- bin2[n]
          starti <- ifelse(i == 1, 1, i - 1)
          endi <- ifelse(i == size, i, i + 1)
          startj <- ifelse(j == 1, 1, j - 1)
          endj <- ifelse(j == size, j, j + 1)
          
          ## upregulated value
          x1 <-
            x2 <- seq(Lambda[n, g][1] - 10, Lambda[n, g][1] + 10, length = 100)
          dens <- matrix(dmvnorm(
            expand.grid(x1, x2),
            sigma = rbind(c(6, 0), c(0, 6)),
            mean = c(Lambda[n, g][1], Lambda[n, g][1])
          ),
          ncol = length(x1))
          dens <- dens * Lambda[n, g][1] / max(dens)
          
          ## downregulated valuee
          x1 <-
            x2 <- seq(Lambda[n, !g][1] - 10, Lambda[n, !g][1] + 10, length = 100)
          dens2 <- matrix(dmvnorm(
            expand.grid(x1, x2),
            sigma = rbind(c(1, 0), c(0, 1)),
            mean = c(Lambda[n, !g][1], Lambda[n, !g][1])
          ),
          ncol = length(x1))
          dens2 <- dens2 * Lambda[n, !g][1] / max(dens2)
          #print(paste("(",i,",",j,")","width=",endj- startj+1,"height",endj-startj+1))
          for (n1 in starti:endi) {
            n1pos <- 50 + 3 * (n1 - i)
            for (n2 in startj:endj) {
              n2pos <- 50 + 3 * (n2 - j)
              name <- paste(n1, "_", n2, collapse = "", sep = "")
              pos <- match(name, object$gene.names)
              new.val <- dens[n1pos, n2pos]
              new.val2 <- dens2[n1pos, n2pos]
              ## if it already exists just modify it
              if (!is.na(pos) > 0) {
                Lambda[pos, g] <- new.val
                Lambda[pos, !g] <- new.val2
                if (!pos %in% ind) {
                  ind <- c(ind, pos)
                  Tags <- nTags + 1
                }
              }
              ## otherwise add it
              else{
                val <- rep(0, ncol(Lambda))
                val[g] <- new.val
                val[!g] <- new.val2
                Lambda <- rbind(Lambda, val)
                object$gene.names <- c(object$gene.names, name)
                ind <- c(ind, nrow(Lambda))
                nTags <- nTags + 1
              }
            }
          }
          
        }
        
        #Lambda <- t(t(Lambda)/colSums(Lambda))
        object$Lambda <- Lambda
        object$indDE <- ind
        object$indNonDE <- (1:nTags)[-ind]
        object$nTags <- nTags
        object$mask_DEup <-
          object$mask_DEdown <-
          object$mask_nonDE <- expandAsMatrix(FALSE, dim = dim(Lambda))
        object$mask_DEup[ind[fcDir == 1], g] <- TRUE
        object$mask_DEup[ind[fcDir == -1], !g] <- TRUE
        object$mask_DEdown[ind[fcDir == -1], g] <- TRUE
        object$mask_DEdown[ind[fcDir == 1], !g] <- TRUE
        object$mask_nonDE[-ind, ] <- TRUE
        object$mask_DE <- object$mask_DEup | object$mask_DEdown
      }
      if (foldDiff == 1 | pDiff == 0)
        object$indDE <- NA
      object
    }
    sim.fun <- function(object)
    {
      ## it is low-level function of NBsim ##
      ## it simulate counts using rnbinom ##
      Lambda <- object$Lambda
      Dispersion <- object$Dispersion
      nTags <- object$nTags
      nlibs <- object$nlibs
      lib.size <- object$lib.size
      counts <-
        matrix(
          rnbinom(
            nTags * nlibs,
            mu = t(t(Lambda) * lib.size),
            size = 1 / Dispersion
          ),
          nrow = nTags,
          ncol = nlibs
        )
      rownames(counts) <-
        object$gene.names  #paste("ids", 1:nTags, sep = "")
      object$counts <- counts
      object
    }
    
    outlier.fun <-
      function(object,
               outlierMech,
               pOutlier,
               min.factor = 2,
               max.factor = 5)
      {
        ## it is low-level function of NBsim ##
        ## it makes outlier ##
        outlierMech <- match.arg(outlierMech, c("S", "M", "R"))
        dim <- dim(object$counts)
        outlier.factor <- function()
          runif(1, min.factor, max.factor)
        countAddOut <- object$counts
        LambdaAddOut <- object$Lambda
        DispersionAddOut <- object$Dispersion
        switch(outlierMech,
               S = {
                 mask_outlier <- expandAsMatrix(FALSE, dim = dim)
                 id_r <- which(runif(dim[1]) < pOutlier)
                 id_c <- sample(dim[2], length(id_r), replace = TRUE)
                 for (i in seq(id_r))
                   mask_outlier[id_r[i], id_c[i]] <- TRUE
                 countAddOut[mask_outlier] <-
                   sapply(countAddOut[mask_outlier], function(z)
                     round(z * outlier.factor()))
               },
               R = {
                 mask_outlier <-
                   matrix(runif(dim[1] * dim[2]) < pOutlier, dim[1], dim[2])
                 countAddOut[mask_outlier] <-
                   sapply(countAddOut[mask_outlier], function(z)
                     round(z * outlier.factor()))
               },
               
               M = {
                 mask_outlier <-
                   matrix(runif(dim[1] * dim[2]) < pOutlier, dim[1], dim[2])
                 LambdaAddOut[mask_outlier] <-
                   sapply(LambdaAddOut[mask_outlier], function(z)
                     z * outlier.factor())
                 countAddOut[mask_outlier] <-
                   rnbinom(
                     sum(mask_outlier),
                     mu = t(t(LambdaAddOut) * object$lib.size)[mask_outlier],
                     size = 1 / DispersionAddOut[mask_outlier]
                   )
               })
        if (!object$foldDiff == 1 & !pDiff == 0)
        {
          indDEupOutlier <-
            which(apply(object$mask_DEup & mask_outlier, 1, any))
          indDEdownOutlier <-
            which(apply(object$mask_DEdown & mask_outlier, 1, any))
          indDEnoOutlier <-
            which(apply((object$mask_DE & !mask_outlier) , 1, all))
          indNonDEOutlier <-
            which(apply(object$mask_nonDE & mask_outlier, 1, any))
          indNonDEnoOutlier <-
            which(apply((object$mask_nonDE & !mask_outlier) , 1, all))
          indDEbothOutlier <- NA
          o <- indDEupOutlier %in% indDEdownOutlier
          q <-  indDEdownOutlier %in% indDEupOutlier
          if (any(o))
          {
            indDEupOutlier <- indDEupOutlier[!o]
            indDEbothOutlier <- indDEupOutlier[o]
            indDEdownOutlier <- indDEdownOutlier[!q]
          }
        }
        else
        {
          indDEupOutlier <-
            indDEdownOutlier <-
            indDEnoOutlier <-
            indNonDEOutlier <- indNonDEnoOutlier <- indDEbothOutlier <- NA
        }
        out <-
          list(
            countAddOut = countAddOut,
            outlierMech = outlierMech,
            pOutlier = pOutlier,
            mask_outlier = mask_outlier,
            indDEupOutlier = indDEupOutlier,
            indDEdownOutlier = indDEdownOutlier,
            indDEbothOutlier = indDEbothOutlier,
            indDEnoOutlier = indDEnoOutlier,
            indNonDEOutlier = indNonDEOutlier,
            indNonDEnoOutlier = indNonDEnoOutlier,
            LambdaAddOut = LambdaAddOut,
            DispersionAddOut = DispersionAddOut
          )
        
      }
    
    calProb <-
      function(x, l)
        round(1 - (1 - x) ^ (1 / l), 2) ## calculate probability to make sure all the outlierMech produce the same amount of outliers ##
    
    
    if (verbose)
      message("Preparing dataset.\n")
    if (class(dataset) == "DGEList")
    {
      dat <- dataset
      dat[["R"]] <-
        dat[["S"]] <-
        dat[["M"]] <- dat[["pOutlier"]] <- dat[["outlierMech"]] <- NULL
    }
    else if (is.character(dataset))
    {
      load(dataset)
      dat <- get(gsub("(\\.)(\\w+)", "", basename(dataset)))
      dat[["R"]] <-
        dat[["S"]] <-
        dat[["M"]] <- dat[["pOutlier"]] <- dat[["outlierMech"]] <- NULL
    }
    else if (is.matrix(dataset))
    {
      if (is.null(name))
        name <- deparse(substitute(dataset))
      geneNames = rownames(dataset)
      dataset <-
        .getDataset(
          counts = dataset,
          drop.extreme.dispersion = drop.extreme.dispersion,
          drop.low.lambda = drop.low.lambda
        )
      dat <-
        new(
          "DGEList",
          list(
            dataset = dataset,
            nTags = nTags,
            lib.size = lib.size,
            nlibs = nlibs,
            group = group,
            design = model.matrix( ~ group),
            pDiff = pDiff,
            pUp = pUp,
            foldDiff = foldDiff,
            outlierMech = outlierMech,
            min.factor = min.factor,
            max.factor = max.factor,
            name = name,
            gene.names = geneNames
          )
        )
    }
    else
      dat <-
      new(
        "DGEList",
        list(
          dataset = dataset,
          nTags = nTags,
          lib.size = lib.size,
          nlibs = nlibs,
          group = group,
          design = model.matrix( ~ group),
          pDiff = pDiff,
          pUp = pUp,
          foldDiff = foldDiff,
          outlierMech = outlierMech,
          min.factor = min.factor,
          max.factor = max.factor,
          name = name
        )
      )
    
    if (!only.add.outlier)
    {
      if (is.null(lib.size))
        dat$lib.size <-
          runif(
            nlibs,
            min = 0.7 * median(dat$dataset$dataset.lib.size),
            max = 1.3 * median(dat$dataset$dataset.lib.size)
          )
      
      if (is.null(nTags))
        dat$nTags <- dat$dataset$dataset.nTags
      if (verbose)
        message("Sampling.\n")
      dat <- sample.fun(dat)
      if (verbose)
        message("Calculating differential expression.\n")
      dat <- diff.fun(dat)
      if (verbose)
        message("Simulating data.\n")
      dat <- sim.fun(dat)
    }
    if (add.outlier) {
      outlierMech <-
        match.arg(outlierMech,  c("S", "R", "M"), several.ok = TRUE)
      if (length(pOutlier) == 1 &
          length(outlierMech) > 1 & any(outlierMech == "S"))
      {
        prob <- calProb(pOutlier, length(group))
        pOutlier <- rep(pOutlier, length = length(outlierMech))
        pOutlier[!outlierMech == "S"] <- prob
      }
      else if (!length(pOutlier) == length(outlierMech))
        stop("pOutlier is not equal to outlierMech")
      if (verbose)
        message("Adding outliers.\n")
      dat[outlierMech] <-
        mapply(
          function(x, y)
            outlier.fun(
              dat,
              outlierMech = x,
              pOutlier = y,
              min.factor = min.factor,
              max.factor = max.factor
            ),
          x = outlierMech,
          y = pOutlier,
          SIMPLIFY = FALSE
        )
      dat$pOutlier <- pOutlier
    }
    
    if (save.file)
    {
      ## save file for shiny app ##
      if (verbose)
        message("Saving file.\n")
      if (is.null(file))
      {
        g <-
          paste0("g", sum(levels(group)[1] == group), "v", sum(levels(group)[2] == group))
        f <- paste0("f", foldDiff)
        if (add.outlier)
          o <- paste0("o", sprintf("%02d", 100 * pOutlier[1L]))
        else
          o <- paste0("o", sprintf("%02d", 0))
        file <- paste0(dat$name, "/", g, f, o, ".Rdata")
        dir.create(dat$name, showWarnings = FALSE)
      }
      filenm <- eval(gsub("(\\.)(\\w+)", "", basename(file)))
      assign(filenm, dat)
      save(list = filenm, file = file)
    }
    dat
  }






#' generate a diffrential chromatin simulation
#'
#' This function uses a binomial distribution to generate the counts of the
#' simulated samples.
#' Additionally, for the differentially intracting regions, we use a gaussian
#' smoothing function to elumate the effect of neighbors change.
#' The implementation of this function is mainly based on the code by at :
#'  http://130.60.190.4/robinson_lab/edgeR_robust/robust_simulation.R
#' with modifications to add the gaussian smoothing part.
#'
#' @param baseHic a dgCMatrix used a base to generate the simulated
#' diffrential Hi-C matrices. The values in the \Rcode{baseHic} represent
#' the interaction means.
#' @param pUp numeric. Probability that a diffrential interaction will show an
#'  increase in frequency in the second condition.
#' The probability to show a decrease is equal to \Rcode{1-pUp}.
#' @param foldDiff double. The amout of fold change.
#' @param pDiff double. The probability that an interaction will be diffrential.
#' @param nrep1 numeric. The number of replicates in the first condition.
#' @param nrep2 numeric. The number of replicates in the second condition.
#' @param extreme.dispersion double. Remove interactions that show
#'  a dispersion value larger than this value.
#'
#' @return a list that contains the simulates replicates and the
#' matrix of the true DCI regions.
#' The list contains the following elements:
#' Hic1 : A list containing the simulated Hi-C matrices of the first condition.
#' Hic2 : A list containing the simulated Hi-C matrices of the second condition.
#' trueStates : A sparceMatric containing the position of the diffrential interactions.
#' @export
#'
#' @examples
#' \dontrun{
#' require(gridExtra)
#' data(K562_matrix_small)
#' simRes <- generateSimulation(K562_matrix_small[1:300,1:300],pDiff = 0.01,pUp = 0.4,
#'                              foldDiff = 5,extreme.dispersion = 0.1,
#'                              nrep1=2,nrep2=2)
#' # Display the simulated data
#' p1 <- levelplot(as.matrix(simRes$Hic1[[1]]), par.settings=YlOrRdTheme,main="Cond1 Rep1")
#' p2 <- levelplot(as.matrix(simRes$Hic1[[2]]), par.settings=YlOrRdTheme,main="Cond1 Rep2")
#' p3 <- levelplot(as.matrix(simRes$Hic2[[1]]), par.settings=YlOrRdTheme,main="Cond2 Rep1")
#' p4 <- levelplot(as.matrix(simRes$Hic2[[2]]), par.settings=YlOrRdTheme,main="Cond2 Rep2")
#'
#' #Plot the position of the diffrentlly interacting regions
#' p5 <- levelplot(as.matrix(simRes$trueStates), par.settings= YlOrRdTheme, main="DCIs position")
#' ## Plot the log2 Fold-change between the two conditions
#' m1 <- simRes$Hic1[[1]] + simRes$Hic1[[2]]
#' m2 <- simRes$Hic2[[1]] + simRes$Hic2[[2]]
#' p6 <- levelplot(as.matrix(log2((m1+0.1)/(m2+0.1))), par.settings = BuRdTheme, main="log2(Cond1/cond2)")
#' grid.arrange(p1,p2,p3,p4,p5,p6, ncol=3)
#' }
setMethod("generateSimulation", signature = c(
  baseHic = "dgCMatrix",
  pUp = "numeric",
  foldDiff = "numeric",
  pDiff = "numeric",
  nrep1 = "numeric",
  nrep2 = "numeric",
  extreme.dispersion = "numeric"
),
function(baseHic = NULL,
         pUp,
         foldDiff = 20,
         pDiff = 0.01,
         nrep1 = 2,
         nrep2 = 2,
         extreme.dispersion = 0.1) {
  smry <- summary(baseHic)
  
  smry <- subset(smry, !is.na(x))
  
  if (nrow(smry) == 0) {
    stop("It seems that the provided base-matrix in empty")
  }
  
  npos = nrow(smry)
  
  counts <- matrix(0, nrow = npos, ncol = nrep1 + nrep2)
  
  ## generate initial interaction values.
  for (i in 1:npos) {
    counts[i, 1:nrep1] <-
      rnbinom(nrep1, mu = round(smry$x[i]), size = 1 / 0.001)
    counts[i, (nrep1 + 1):ncol(counts)] <-
      rnbinom(nrep2, mu =  round(smry$x[i]), size = 1 / 0.001)
  }
  rownames(counts) <-
    paste0(smry$i[1:npos], "_", smry$j[1:npos])
  
  
  group <- factor(c(rep(1, nrep1), rep(2, nrep2)))
  
  ## Generate diffrentially expressed interactions
  simData <-
    .NBSim.neig(
      dataset = counts,
      group = group,
      nTags = npos,
      pUp = pUp,
      foldDiff = foldDiff,
      drop.extreme.dispersion = extreme.dispersion,
      pDiff = pDiff,
      add.outlier = FALSE
    )
  
  
  ## Get the position of the interactions
  matPosi <-
    as.numeric(gsub("_\\d+", "", rownames(simData$counts)))
  matPosj <-
    as.numeric(gsub("\\d+_", "", rownames(simData$counts)))
  
  
  ## Fill in the matrices of the first condition
  Hic1 <- list()
  pos <- which(group == 1)
  
  for (i in 1:length(pos)) {
    p = pos[i]
    Hic1[[i]] <- sparseMatrix(
      i = matPosi,
      j = matPosj,
      x = simData$counts[, p],
      dims = dim(baseHic)
    )
  }
  
  
  ## Fill in the matrices of the second condition
  Hic2 <- list()
  pos <- which(group == 2)
  
  for (i in 1:length(pos)) {
    p = pos[i]
    Hic2[[i]] <- sparseMatrix(
      i = matPosi,
      j = matPosj,
      x = simData$counts[, p],
      dims = dim(baseHic)
    )
  }
  
  
  is1 <-
    as.numeric(gsub("_\\d+", "", simData$gene.name[simData$indDE]))
  js1 <-
    as.numeric(gsub("\\d+_", "", simData$gene.name[simData$indDE]))
  trueStates <- sparseMatrix(
    i = is1,
    j = js1,
    x = rep(1, length(is1)),
    dims = dim(baseHic)
  )
  
  
  simParams = list(
    pUp = pUp,
    foldDiff = foldDiff,
    pDiff = pDiff,
    nrep1 = nrep1,
    nrep2 = nrep2,
    extreme.dispersion = extreme.dispersion
  )
  
  res <- new(
    "SimulatedHiC",
    Hic1 = Hic1,
    Hic2 = Hic2,
    trueStates = trueStates,
    params = simParams
  )

  return(res)
})



setMethod("show","SimulatedHiC",
          function(object){
            cat("SimulatedHiC object\n")
            
            if( length(object@Hic1)>0 & length(object@Hic2) >0){
              cat(paste("Condition 1 with", length(object@Hic1),"replicates\n"))
              cat(paste("Condition 2 with", length(object@Hic2),"replicates\n"))
              
              nrows = nrow(object@Hic1[[1]])
              ncols = ncol(object@Hic1[[2]])  
              cat(paste0("Simulated matrix size (",nrows,",",ncols,")\n"))
            }
            
          }
          )
