getNorm <- function(replicates = list(), plt=FALSE,fout="M-A plot.png"){
  
  require(limma)
  if(length(replicates) < 2)
    stop("At least two replicates should be provided")
  
  nbBins <- dim(replicates[[1]])
  
  smry1 <- summary(replicates[[1]])
  smry2 <- summary(replicates[[2]])
  
  mrg <- merge(smry1, smry2, by=c("i","j"), all=TRUE, suffixes = c(1,2))
  rm(smry1,smry2)
  
  mrg$x1[is.na(mrg$x1)] <- 0
  mrg$x2[is.na(mrg$x2)] <- 0
  mrg$x1 <- mrg$x1+1
  mrg$x2 <- mrg$x2+1
  
  M <- log2( mrg$x1 / mrg$x2)
  A <- 0.5 * log2(mrg$x1 * mrg$x2)
  
  o <- order(A)
  A <- A[o]; M <- M[o]
  
  fit <- loessFit(M,A, iterations = 5)
  
  if(plt){
    message(paste("Plotting M-A plot into file",fout))
    png(fout, width=1024, height=1024)
    plot(A,M, main="M-A plot between replicates")
    lines(A,fit$fitted, col='red',lwd=2)
    dev.off()
  }
  
  
  mrg <- mrg[o,]
  
  M_correct = M- fit$fitted
  
  replicates[[1]] <- sparseMatrix(i= mrg$i, j=mrg$j, x= 2^(A + 0.5 * M_correct),dims=nbBins)
  replicates[[2]] <- sparseMatrix(i= mrg$i, j=mrg$j, x= 2^(A - 0.5 * M_correct),dims=nbBins)
  return(replicates)
}



.adjustPvalueByDiagonal <- function(res){
  
  require(dplyr)
  
  mrg <- summary(res)
  #mrg <- subset(mrg, x > 0)
  mrg <- subset(mrg, i < j)
  
  # calculate the distance to group them by
  mrg$dist <- mrg$j - mrg$i
  # calculate the FDR diagonal-wise
  res2 = mrg %>% group_by(dist) %>% mutate(padj=p.adjust(p=x,method = "BH"))
  
  return(res2)
}



#' .hardCutof: Get significant DCI using hardcutoff 
#'
#' @param res a dgCMatrix matrix
#' @param isrOP_qval logical, indicating where or not to use the provided qvalue
#'  as cutoff value (FALSE) or use the rOP estimated p-value as cutoff (TRUE).
#'  Here the rOP p-value is estimated on a window sized: windowSize * windowSize, 
#'  however, it will be from the rOP p-value for thepairwise interactions 
#'  located in the borders, or interactions located  in the nearby or repeat 
#'  regions as they have a smaller window size.
#
#' @param windowSize integer, (default 3), indicates the size of the sliding 
#' window
#' @param alpha double, (default 0.7), only used if isrOP_qval==TRUE. 
#' It indicates the percentage of p-values that should be considered significant
#' in a given window
#' @param qvalue the hard cutof qvalue to use. If isrOP_qval==TRUE, the 
#' q-value will be euqal to : `pbeta(qvalue, r, ceiling(alpha * 2 * windowSize ^ 2) - r + 1)`
#'
#' @return
#'  a data.frame with three columns : i,j,x and padj that indicate the 
#'  interaction (i,j) and the pvalue and the qvalue.
#' @export
#'
#' @examples
.hardCutoff <- function(mrg, isrOP_qval=FALSE, windowSize=3, alpha=0.7, qvalue=0.01) {
  
  if (isrOP_qval) {
    nbTests <- 2 * windowSize ^ 2
    r = ceiling(alpha * nbTests)
    fdr <- pbeta(qvalue, r, nbTests - r + 1)
    mrg_sig <- subset(mrg, padj < fdr)
    
  } else{
    mrg_sig <- subset(mrg, padj < qvalue)
  }
  
  return(mrg_sig)
}



#' .quantRegressionthereshold: Get significant DCI using quantile regression.
#'
#' @param chrDCI 
#' @param windowSize 
#' @param quantile 
#'
#' @return
#' return the quantile-regression filted qvalues according to the interaction 
#' span
#' @export
#'
#' @examples
.quantRegressionthereshold <- function(mrg, windowSize, quantile) {
  
  mrg$dist <- mrg$j - mrg$i
  mrg$padj <- -log10(mrg$padj)
  nbTests <- 2 * windowSize ^ 2
  #r = ceiling(alpha * nbTests)
  # generally values corresponding to alpha=0.01 are very dominant
  #mrg <-subset(mrg , x > -log10(pbeta(0.01, r, nbTests - r + 1)))
  
  o <- order(mrg$dist)
  mrg <- mrg[o, ]
  X <- model.matrix(mrg$padj ~ splines::bs(mrg$dist, df = 9))
  
  thr  <-
    quantreg::rq(padj ~ splines::bs(dist, df = 9),
                 data = mrg,
                 tau = quantile)
  thr.fit <- X %*% thr$coef
  mrg$threshold <- thr.fit
  
  mrg_sig <- subset(mrg, padj > threshold)
  
  return(mrg_sig)
}


.getclusteredRegions <- function(res, windowSize, alpha){
  # Get the clustered regions that have at least 2*windowSize^2
  res[res==1] = 0
  x = res %>% 
      as.matrix %>% 
      as.cimg %>% 
      threshold(0) 

  lab <- label(x)
  lab <- as.matrix(lab * as.cimg(x))
  
  tbl = table(lab)
  max = names(tbl)[which.max(tbl)]
  max = as.numeric(max)
  
  res[lab==max] = 0

  res
}

